"""This example should illustrate the use of self-referential descent production.
The way p-values are references is not very comfortable, but it allows to access arbitrary
p-siblings through a very simple core-implementation.  
"""
from project_conf import *

from peewit.core import ENode, Experiment
from peewit.presentation import Viewer 
from cut.baseclass import Object

from ml.table import TableData
from ml.ml_data import ionosphere

from ml.perceptron import SimplePerceptron
from ml.kernel import  Poly



class data(ENode):
    
    def init(self, build=True, store=False):
        pass
    
    def descent(self):
        ## the user uses pathkeeper to get the path to the data
        TableData.global_datadir = PK.abs_path('example_data_sets/')
        ##  some setting in users code
        TableData.global_build   = self.build
        TableData.global_store   = self.store
        

        return  [
                  ionosphere(build=1),
                ]
        




class algorithm(ENode):
    """values expected to have train and predict method"""

    def descent(self, data, kernel):
        
        
        engine = SimplePerceptron( data_tray   = data,
                                   kernel      = kernel,
                                   n_epoches   = 16 
                                  )
                

        return [ engine ]
    
    
class training_size(ENode):
    
    def descent(self):
        
        R = [  10, 20, 30, 40, 50 ]
        return R
 

class kernel(ENode):
    
    def descent(self):
        
        k_1 = Poly(g=2)
        k_2 = Poly(g=3 )
        
        return  [ k_1,
                  k_2  
                  ]
  

class leaf(ENode):


    def descent(self, algorithm, data, sample, training_size, leaf):
        
        A, E   = sample
        
        A      = A[:training_size]
        
        # get the model obtained from viewer examples
        alpha_previous = leaf.training_size
        
        alpha, S  = algorithm.train(A, alpha=alpha_previous)
        
        Y      = data.get_YS(E)
        
        Y_hat  = algorithm.predict((alpha,S), E) 
        
        accuracy   =  1. * sum(  Y == Y_hat ) / sum( ( Y == 1 ) + (Y == -1) )
        
        ## p-node will store that value and it will become another alpha_previous
        self.svalue = alpha
        
        return [  accuracy  ]
        
    
    def initial_siblings(self):
        """Set the initial sibling values as dict that maps ancestor e-node to initial s-values.
        """    

        # only one sibling is queried by self.descent        
        # and it expects a list as sibling value 
        out = Object()
        out.training_size = []
        
        return out
  

from example_1 import sample


etree = [     data(),
              sample(), 
              training_size(),
              kernel(),
              algorithm(), 
              leaf(),
        ]
      
    
    
if __name__ == '__main__':


    E = Experiment(pathkeeper=PK,
                   version   = 10
                   )

    
    E.define(etree)

    vc = E.run()
    
    
    vi = Viewer(vc)
    
    vi.print_vcube()


