
from cut.baseclass import Named, Object
from cut.numpytools import indices_equal, quantiles
from cut.container import indices

from ml.preprocessing import Standardization

from numpy import array, shape, sum, empty, linalg
from random import shuffle




class BinaryNaiveBayes(Named): 
    """Generic Naive Bayes Training for binary output and
    binary features."""
    
    def init(self,
                data_tray   = None,
        ):
        
        pass
    
    def train(self, A):
        
        Phi = self.data_tray.get_X(A)
        Y   = self.data_tray.get_YS(A)
                
        assert shape(Phi)[0] == len(Y) and len(shape(Y)) == 1
        assert set(Y) == set([-1,+1])  and set(Phi.ravel()) == set([0,1])

        I_n = indices_equal(Y,-1)
        I_p = indices_equal(Y,+1)
        
        n_n = len(I_n) * 1.
        n_p = len(I_p) * 1.
        
        Phi_n = Phi[I_n]
        Phi_p = Phi[I_p]

        m = Object() # model
        
        m.w_n_n = sum(Phi_n == 0, axis=0)/n_n
        m.w_n_p = sum(Phi_n == 1, axis=0)/n_n
        m.w_p_n = sum(Phi_p == 0, axis=0)/n_p
        m.w_p_p = sum(Phi_p == 1, axis=0)/n_p
        
        m.b_n   = n_n/(n_n + n_p)
        m.b_p   = n_p/(n_n + n_p)
        
        
        return m      
        
    
    def predict(self,model, T ):
        Phi = self.data_tray.get_X(T)
        
        assert len(shape(Phi)) == 2 and set(Phi.ravel()) == set([0,1])
        m = model
        Y_hat = []
        for phi in Phi:
            
            phi_n  = ( phi == 0 )
            phi_p  = ( phi == 1 )


            f_n    = sum(m.w_n_n * phi_n + m.w_n_p * phi_p) + m.b_n 
            f_p    = sum(m.w_p_n * phi_n + m.w_p_p * phi_p) + m.b_p
            
            Y_hat += [ (f_n < f_p) * 2 - 1 ]
             
        return Y_hat



class PhiQuantiles(Named):
    """Binary Feature building"""
    
    @staticmethod
    def build(X_A, X_B, Y_A, k=2):
        
        x     = X_A[0]
        J_all = indices(x)
        J_b   = []
        S_bmax = {}
        Q      = {}
        for j in J_all:
            C = X_A[:,j]
            if len(set(C)) <= 2:
                J_b += [ j]
                S_bmax[j] = max(C) 
            else:
                Q[j] = [0.] + quantiles(C,k)
                
        def PhiG(X):   
            Phi = []
            for x in X:
                
                phi = []
                for j in J_all:
                    if j in J_b:
                        phi += [ int(x[j] == S_bmax[j]) ]
                    else:
                        for l in range(1,k+1):
                            phi += [ int((x[j] <= Q[j][l]) * (Q[j][l-1] < x[k])) ]
                Phi += [phi]
            return array(Phi)
        
        if not X_B is None:
            return PhiG(X_A),PhiG(X_B)
        else:
            return PhiG(X_A)
        
        
class PhiNearerNeighboor(Named):
    """Binary Feature building"""
    
    @staticmethod
    def build(X_A, X_B, Y_A,k=None):
        
        pp = Standardization()
        
        pp.train(X_A)
        X_A = pp.apply(X_A)
        X_B = pp.apply(X_B)
        
        n_A,d     = shape(X_A)
        n_B,d     = shape(X_B)
       
        D_A  = empty((n_A,n_A))
        D_BA = empty((n_B,n_A))
        
        for i in range(n_A):
            for j in range(n_A):
                D_A[i,j] = linalg.norm(X_A[i]-X_A[j])
        
        
        #assert all(all(D_A == D_A.transpose()))

        for i in range(n_B):
            for j in range(n_A):
                D_BA[i,j] = linalg.norm(X_B[i]-X_A[j])
        

        I_n = indices_equal(Y_A,-1)
        I_p = indices_equal(Y_A,+1)
        
        shuffle(I_n)
        shuffle(I_p)
        
        assert len(I_n) >= 40
        assert len(I_p) >= 40
        
        pp_1 = I_p[00:10]
        pp_2 = I_p[10:20]
        pn_1 = I_p[20:30]
        np_2 = I_p[30:40]
        nn_1 = I_n[00:10]
        nn_2 = I_n[10:20]
        np_1 = I_n[20:30]
        pn_2 = I_n[30:40]
        
        P = []
        
        for i in range(10):
            P += [ (pp_1[i],pp_2[i]) ]
            P += [ (nn_1[i],nn_2[i]) ]
            P += [ (pn_1[i],pn_2[i]) ]
            P += [ (np_1[i],np_2[i]) ]
        
        d_ = len(P)
        
        Phi_A = empty((n_A,d_))
         
        for i in range(n_A):
            for j in range(d_):
                Phi_A[i,j] = (D_A[i,P[j][0]] < D_A[i,P[j][1]])

        Phi_B = empty((n_B,d_))
         
        for i in range(n_B):
            for j in range(d_):
                Phi_B[i,j] = (D_BA[i,P[j][0]] < D_BA[i,P[j][1]])

        Phi_A = Phi_A.astype(int)
        Phi_B = Phi_B.astype(int)
        
        return Phi_A,Phi_B        
                
        
        
    

if __name__ == '__main__':
    
    pass    
