from numpy import dot,  linalg, exp, shape, mean, diag ,repeat,\
                  sqrt, nanmax, ones, empty, reshape, nan, array, \
                  log2, median, random,max, min, nanmin, argmin, zeros, isnan


from cut.baseclass import Named
from cut.numpytools import  tp, col, row, quantile
from collections import OrderedDict
try:
    from cu.plotting import FunctionPlot
    import pylab
except:
    pass

class Kernel(Named):
     
    def init(self,
             c     = 0,
             g     = 1,
             sigma = 1,
             fixed = False,
             inner = None,
             prep  = None,
             posp  = None,
             normalize=1,
             ):
        
        if self.has('apply_outer'):
            assert self.inner
    
    
    def apply_crosswise(self, X1, X2):
        """f_ is float valued"""
        n_1 = len(X1)
        n_2 = len(X2) 
        out = nan * empty((n_1,n_2), dtype='float32') 
        
        for i in range(n_1):
            for j in range(n_2):
                out[i,j] = self.apply(X1[i],X2[j])
                if self.normalize:
                    out[i,j] = out[i,j]/sqrt(self.apply(X1[i],X1[i]) * 
                                             self.apply(X2[j],X2[j]))
        return out

 
class PosP(Named):    
    
    def train(self, K):
        pass
    
    def apply(self, K):
        return K
    
class Mikimaki(PosP):
    
    def train(self, K):
        K1 = K - diag(nan*diag(K))
        self.miki = nanmin(K1.ravel())
        K2 = K - diag(diag(K))
        self.maki = nanmax(K1.ravel())
        self.maki = quantile(K.ravel(), 1 - 1./8)
        #self.send("miki maki", self.miki, self.maki)
    
    def apply(self, K):
        #K =  (K - self.miki) / self.maki    
        K =  K / self.maki
        return K


class DotProduct(Kernel):

    def apply(self,x1,x2):
        return dot(row(x1), col(x2)).ravel()[0]
    
    def apply_crosswise(self,X1, X2):
        return dot(X1, tp(X2))


class Diff1Norm(Kernel):
    
    def apply(self, x1, x2):
        d = len(x1)
        return 1. * linalg.norm( x1-x2, ord=1)#/ d


class Diff2Norm(Diff1Norm):
    
    def apply(self, x1, x2):
        d = len(x1)
        return 1. * linalg.norm( x1-x2, ord=2) * 1. / sqrt(d)
    

class DiffASQ(Diff1Norm):
    
    def apply(self, x1, x2):
        d = len(x1)
        return  2**(sum((abs(x1-x2))) * 1. / d )
        #return 1. * linalg.norm( x1-x2, ord=2) #/ sqrt(d)
    
    
    

class Poly(Kernel):

    def init(self,
                c = 0.,
                g = None,
                inner=DotProduct()):
        
        pass
    
    def apply_outer(self, T):    
        return ( T + self.c )**self.g
    
    
    def apply(self, x, x_):
        return ( self.inner.apply(x, x_) + self.c )**self.g

        

class Linear(Poly):
    
    def init(self,
             c    = 0.,
             g    = 1.
             ):
        assert c == 0.
        assert g == 1.
        


class Exp1(Kernel):
    def init(self,
             sigma = 1.,
             inner = Diff1Norm(),
             ):
        pass
        
    def apply_outer(self, T):    
        return exp( - (T / self.sigma ) )
    
    def apply(self, x,x_):
        return exp( - self.inner.apply(x-x_) / ( self.sigma)  )



class Exp2(Exp1):
    def init(self,
             inner  = Diff2Norm()
             ):
        pass
         

class Gauss(Kernel):

    def init(self,
             sigma = 1.,
             inner  = Diff2Norm() ):
        pass

    def apply_outer(self, T):    
        return exp( - (T**2 / ( 2 * self.sigma) ) )
        
    def apply(self, x,x_):
        return exp( - (self.inner.apply(x-x_)**2 / ( 2 * self.sigma) ) )




class KernelTray(Named):

    def init(self,
             
             kernel=None,
             data_tray=None,
             
             use_cache = False,
             cache={},
             
             pruning_log_bound=None,
             q_fix       = None,
             t_fix       = None,
             
             precomputed = False,
             use_precomputed = False,
             
             
             ):
        pass

    def precompute(self,A,B):
        self.A_pre = A
        self.B_pre = B
        self.K_pre = self.query(A, B)
        self.precomputed = True
        

        
    def query(self, A,B, kernel=None ):
        """
        Return kernel-values for all index-pairs AxB in matrix form. 
        Implements caching, the values at are provided by self._query 
        at first place. 
        """
        
        
        
        if not kernel: 
            kernel = self.kernel
        
        
        if self.use_precomputed and self.precomputed:
            iA = [ self.A_pre.index(a) for a in A]
            iB = [ self.B_pre.index(b) for b in B]
            return self.K_pre[iA,:][:,iB]
        
        
        if self.use_cache: 
            
            key_kernel = self.kernel_hash(kernel)
            
            key_index  = ( tuple(A), tuple(B) )
            
            if key_kernel in self.cache:
    
                kernel_instance = self.cache[key_kernel]
                reuse           = False
                    
                for key_index_,K_ in list(kernel_instance.items()):
                    
                    A_, B_  = key_index_
                            
                    if  set(A).issubset(set(A_)) and\
                        set(B).issubset(set(B_)):
                        
                        reuse = True
                        
                        iA = [ A_.index(a) for a in A]
                        iB = [ B_.index(b) for b in B]
                            
                        K  = K_[iA,:][:,iB]
                        break
                    
                if not reuse:
                    K                          = \
                    kernel_instance[key_index] = self._query(A,B, kernel)            
    
            else:
                kernel_instance            = OrderedDict()    
                K                          = \
                kernel_instance[key_index] = self._query(A,B, kernel)           
                self.cache[key_kernel]     = kernel_instance
        else:
            K = self._query(A,B, kernel)  
        
        return K
        
    
    def _query(self,A,B, kernel):
            
        if   kernel.has('apply_outer'):
            KI = self.query(A, B, kernel.inner)
            K  = kernel.apply_outer(KI)
        else:
            XA    = self.data_tray.get_X(A)
            XB    = self.data_tray.get_X(B)
            
            if kernel.prep:
                XA   = kernel.prep.apply(XA)
                XB   = kernel.prep.apply(XB)
            
            assert all( [ not isnan(v) for v in XA.ravel()])
            assert all( [ not isnan(v) for v in XB.ravel()])
            
            K = kernel.apply_crosswise(XA, XB)
            
            
            if kernel.posp:
                K = kernel.posp.apply(K)
        
        return K


    def train_prep(self, kernel, A):
        
    
        if kernel.prep:
        
            X  = self.data_tray.get_X(A)
            
            kernel.prep.training_indexes = A
            
            kernel.prep.train(X)
        
        if kernel.inner.prep:
        
            X  = self.data_tray.get_X(A)
            ## make sure training set is incorporated in kernel-hash 
        
            kernel.inner.prep.training_indexes = A
            
            kernel.inner.prep.train(X)
        
    
        
    
    def train_posp(self, kernel, A):
        if kernel.posp:
            ## make sure training set is incorporated in kernel-hash 
            posp = kernel.posp
            posp.training_indexes = A
            kernel.posp = None
            K = self._query( A, A , kernel=kernel)
            kernel.posp = posp
            kernel.posp.train(K)
        
    
    def kernel_hash(self, k):
        """same hash means same kernel values"""
        
        inner = self.kernel_hash(k.inner) if k.inner else 'no_inner'
        prep  = (k.prep.get_name(),k.prep.training_indexes) if k.prep else 'no_prep'
        posp  = (k.posp.get_name(),k.posp.training_indexes) if k.posp else 'no_posp'
        
        parts = [ k.get_name(), k.c, k.g, k.sigma,
                  inner, prep, posp ]
        
        parts = [ str(p) for p in parts ]

        return "__".join(parts)
    
            
    
    def flash_cache(self):
    
        self.send('flash cache.')          
        self.cache.clear()
       
    
            
    def distances(self, S, T):
        
        #assert not set(S).intersection(T)
        
        K_ST    = self.query(S,T)
        
        K_S  = self.query(S,S)
        K_T  = self.query(T,T)
        
        n = len(S)
        m = len(T)
        
        D    = nan * zeros((n,m)) 

    
        for s in range(n):
            for t in range(m):
                D[s,t] =  sqrt(K_S[s,s] + K_T[t,t] - 2 * K_ST[s,t]) 
                self.send("D:  %.3f = sqrt(%.3f + %.3f - 2*  %.3f )" %\
                          (D[s,t],K_S[s,s] , K_T[t,t] , K_ST[s,t]))
        return D
    



    def decomposition(self,M):
        """Computes eigenvalues decomposition of kernel matrix K.
        Rarely svd fails. Since we do not want this to break the runs we 
        retry with random noise added to K a fixed number of times.
        
        Returns UT,D,U, with orthonormal U. The eigenvectors are rows of U
        and columns of UT. D is one-dimensional.
        """
        
        K = M.K
        n = len(K)
        
        
        success = False
        trials = 0
        
        while not success and trials < 8:
            try:
                QT,D,Q  = linalg.svd(K)
                success = True
            except:
                self.send('SVD failed, retry with altered matrix.')
                K = K + random.rand(n,n) * 2*-20
        if not success:
            raise Exception()
        
        
        ## prune dimensions
        if not self.q_fix is None:
            q = self.q_fix
        elif not self.t_fix is None:
            
            q  = sum( D >= self.t_fix )
            t = self.t_fix
            
        elif not  self.pruning_log_bound is None:
            t  = 2**(mean(log2(D)) + self.pruning_log_bound)
            q  = sum( D >= t )
        
        else:
            q  = len(D)
        
        
        
        if q < n:
            self.send('last-in,first-out:', D[q-1],D[q])
    
        
        self.send("subspace dimensions :  %d/%d" % (q,n))
        
        
        D  = D[:q]
        Q  = Q[:q,:]
        QT = Q[:,:q]
        
        B  = dot(diag( D**(-1./2) ), Q ) 
        
        K_ = dot(tp(B),B)

        S  = sum(1/D)
        
        M.QT, M.D, M.Q, M.B, M.K_, M.S, M.n, M.q = QT,D,Q,B, K_, S, n,q 
         
        return M
    
    
    def km(self,f, v_1, v_2=None):
        
        def f_( x, x_ ):
            #assert f(x,x) >= 0 and f(x_,x_) >= 0, str((f,x, x_))
            return f( x, x_ ) 
        
        if v_2 == None:
            v_2 = v_1
        
        return self.crosswise(f_, v_1, v_2)

    


   

if __name__ == "__main__":

    pass
