"""A libsvm wrapper that uses a hacked libsvm to access model alpha. """

from cut.baseclass import Named
from cut.numpytools import col

from ml.kernel    import  KernelTray, Linear
from ml.model import LLMM
from numpy import array, shape

import svm as libsvm
import svmutil as libsvmutil

class LIBSVM(Named): 
     
    def init(self,
                data_tray   = None,
                kernel      = Linear(),
                C           = 1.,
                nu          = 0.1,
                svm_type    = 'C',
                kt          = None
        ):
        
        
        
        #libsvm.svm_parameter(kernel_type = libsvm.LINEAR)
        
        parameter   = libsvmutil.svm_parameter()#kernel_type = libsvm.LINEAR)
            
        if   svm_type == 'C':
            parameter.svm_type = 0
            parameter.C = C
        elif svm_type == 'nu':
            parameter.svm_type = 1
            parameter.nu = nu
            self.send("nu: ", nu)
        else:
            self.bugger('Wrong svm_type.')
        
        kname = kernel.get_name()
        
        if  kname == 'Linear':    
            parameter.kernel_type = 0

        elif kname == 'Poly':
        
            parameter.kernel_type = 1
            parameter.coef0  = kernel.c
            parameter.gamma  = 1.
            parameter.degree = kernel.g
                
                    
        elif kname == 'Gauss':
            
            parameter.kernel_type = 2
            parameter.gamma = 1./(2. * kernel.sigma)
              
        else:
            self.bugger('Wrong kernel type: ' + kname)
        
        self.parameter = parameter
        
        
    def set_C(self, C, n):
        self.parameter.C = C/float(n) 

    def set_sigma(self,sigma):
        self.sigma = sigma
        self.parameter.gamma = 1./(2. * sigma**2)

    def train(self, S=None, X=None, Y=None, sign=+1):
        """S are indices of the training examples"""
        
        
        if X is None:
            X,Y    = self.data_tray.get_XYS(S)
        
        Y        = sign * Y 
        

        model = LLMM()
        model.sign=sign       
        
        
        model.prep = self.kt.kernel.prep
        
        X_ =    model.prep.apply(X)  
        X_ = X_.tolist()
        
        problem            = libsvmutil.svm_problem(Y, X_)
        model.libsvm_model = libsvmutil.svm_train(problem, self.parameter)
        
        return model
    
    
    def predict(self, model, T=None, X=None):
       
        if X is None:
            X_test            = self.data_tray.get_X(T)
        else:
            X_test = X
        
        
        
        X_test  = model.prep.apply(X_test)
        Y_dummy = [1] * len(X_test)
    
        libsvm_model  = model.libsvm_model
        
        Y_hat  = libsvmutil.svm_predict(Y_dummy,X_test.tolist(), libsvm_model)[0]
        Y_hat  = array(Y_hat).ravel()
        
        return Y_hat



if __name__ == '__main__':
    
    pass    
