
from cut.baseclass import Named
from cut.numpytools import col

from ml.kernel    import Linear

from numpy import array



class SimplePerceptron(Named): 
    """
    An inefficient Rosenblatt-Perceptron that takes preset
    weights as additional input."""
    
    def init(self,
                data_tray   = None,
                kernel      = Linear(),
                n_epoches   = 8,
        ):
        
        pass
           
        
    def train(self, S, alpha=[]):
        """S are indices of the training examples, alpha are preset weights """
        
        X,Y    = self.data_tray.get_XYS(S)

        
        n_new =  len(X) - len(alpha)

        if alpha is None:
            alpha = [ 0 ] * len(X)
        else:
            suffix = [0] * n_new
            alpha = alpha + suffix
        
    
        indices = list(range(len(X)))
        n_updates = 0
        error = True
        round = 0
        while round < self.n_epoches and error:
            
            error = False
            for i in indices:
                if not self.predict( (alpha,S),  [i] )[0] == Y[i]:
                    alpha[i] += 1
                    error = True
                    n_updates += 1
            round += 1
        
        self.send('Training finished, %d updates in total.' % ( n_updates) )
        self.send()                   
        

        
        return alpha, S      
        
    
    def predict(self, model, T):

        alpha, S = model

        X_train, Y_train = self.data_tray.get_XYS(S)        
        
        X_test  = self.data_tray.get_X(T)
        
        Y_hat  = [] 
        
        I_train = list(range(len(X_train)))
        I_test  = list(range(len(X_test)))


        for i in I_test:
            sum = 0.
            for j in I_train:
                sum += self.kernel.apply( X_train[j], X_test[i] ) * alpha[j] * Y_train[j]
                
            Y_hat += [ 1 if sum >= 0 else -1 ]
        
        Y_hat             = array( Y_hat  )
        return Y_hat



if __name__ == '__main__':
    
    pass    
