from numpy import array, shape, log, sum

import re

from cut.baseclass    import Named
from cut.numpytools   import tp

nltk = None
class TextProcessing(Named):


    @staticmethod
    def tokenize_html(raw_input):
        text = TextProcessing.extract_words(raw_input)
        return text

    
    @staticmethod
    def tokenize_html_2(raw_input):
        print("XX1", raw_input)
        text = nltk.clean_html(raw_input)
        return nltk.wordpunct_tokenize(text)

    @staticmethod            
    def extract_words(text):
        return re.findall('[a-z]+', text.lower())
                
    @staticmethod
    def get_terms(documents):
        term_set  = set()
        for document in documents:
            term_set = term_set.union(set(document))
    
        return list(term_set)
    
    
    @staticmethod
    def tf(token_sequence, token_set):
        """Result contains relative frequencies of tokens
        from token_set in token_sequence. Mostly, the sequence represents
        a text document and the token_set the set of term occuring 
        in the training set, hence tf for term frequencies.
        The output vector has shape (m,) where m is the size of token_set."""  
        document = token_sequence
        terms    = token_set
        counts   = array([ document.count(term) for term in terms]) 
        return counts
    
    @staticmethod
    def TF(documents, terms):
        """Batch variant of tf. Can be used to compute idf. Now the
        documents are represented by rows."""
        
        n = len(documents)
        m = len(TextProcessing.tf(documents[0], terms))
        TF = array( [ TextProcessing.tf(doc, terms) for doc in documents ])
        TF = TF.reshape(n,m)
        
        #print "TTT", shape(TF), n, m
        assert shape(TF) == (n,m)
        return TF
        
    @staticmethod
    def IDF(TF):
        """Compute the inverse document frequency vector, that is use
        to weight the entries of a term frequency vector.
        The column of the input matrix TF are assumed to tf vectors of 
        a set of documents.
        The out is a column vector."""
        def h(term_frequency):
            return log( (  float(n_documents ) / (term_frequency + 1) ) ) 
        
        n_terms, n_documents  = shape(TF)
        absolut_df  = sum(TF > 0, 0)
        
        return tp(h(absolut_df))
    

    