"""Functions related to reading and writing files."""

import os, pickle, sys, glob


def load_from_shelve(file_name):
    file = open(file_name, 'rb') 
    o = pickle.load(file)
    return o

def save_in_shelve(o, file_name):
    try:
        file = open(file_name, 'wb')
        pickle.dump(o, file) # always create a new database
    except Exception as e:
        print("save_in_shelve: bugger with %s " % file_name)
        print(e)
        
def save_file(text, path=None, 
              append=False,
              no_change_no_write=False,
              argv_index=1,
              ):
    
    if path is None:
        path = sys.argv[argv_index]
        
    dont = False
    if no_change_no_write:
        if os.path.isfile(path):
            old = read_file(path)
            if text == old:
                dont = True
                
    if not dont:
        mode = 'a' if append else 'w'
        h_out = open(path,mode)
        h_out.write(text)
        h_out.close()


def read_file(path=None, argv_index=1):
    if path is None:
        path = sys.argv[argv_index]
    h_out = open(path,mode='rb')
    out = h_out.read().decode()
    h_out.close()
    return out     

def get_lines(file, comment_suffix=None):
    """Returns list of striped lines from file."""
    fh    = open(file)
    lines = fh.readlines()
    fh.close()
    lines = [ line.strip() for line in lines ]
    if comment_suffix:
        lines = [ line for line in lines 
                  if not line.startswith(comment_suffix)]
    return lines

def get_content_lines(file):
    """Only non-comment lines. Note  that lines are stripped
    before they are checked for comment-suffix."""
    return get_lines(file, comment_suffix="#")



def create_dir(dir_path):
    """Create directories in dir_path, likewise mkdir -P."""
    
    if not os.path.isdir(dir_path):
        head, _ = os.path.split(dir_path)  
        if head:
            create_dir(head)
        os.mkdir(dir_path)
    


def posix_path(path, os_path=os.path):
    """
    @param path: a string that is path descriptor that 
        can be processed by the os_path module. If os_path=os.path,
        the descriptor has to fit to the system on which the function 
        is called.
    @param os_path: the module that is used to interpret path
    @return: the path described by path in posix-notation
    """
            
    elements = split_path(path, os_path=os_path)
    ## shortcut
    if os_path.isabs(path) and not elements[0] is '':
        elements = [''] + elements
    
    return '/'.join(elements)
    
    
    
def split_path(path, os_path=os.path):
    """
    @param path: path descriptor as understood by the module os_path
    @param os_path: the path-processing module used to interpret path 
    @return: list of components of the path. If path is absolute, the 
        first element is ''. If path is a file-path (no ends with a dir),
        the last element is ''.
    """
    prefix, suffix = os_path.split(path)
    
    if prefix == path:
        return [ '' ]
    if prefix:
        splited_prefix = split_path(prefix, os_path=os_path)
        out = splited_prefix + [ suffix ]
    else:
        out = [ suffix ]
    return out



def system_path(*path_descriptors):
    """
    @param path_descriptors: a path_descriptor is path part in posix-notation.
    @return: a string that describes the corresponding path as demanded by the os.
    
    Conterpart of posix_path, but it can take two or more path descriptors 
    that are joint to one path.
       """
               
    if len(path_descriptors) > 1:
        
        head = system_path(path_descriptors[0])
        rargs = path_descriptors[1:]
        rest = system_path(*rargs)

        return os.path.join(head, rest)
    
    else:
        
        path_descriptor = path_descriptors[0]
            
        parts = path_descriptor.split('/')
        
        new_parts = []
        
        part = parts[0]
        if part == '':
            parts.pop(0)
            new_parts += [ os.path.sep ]
        
        for part in parts:
            if part == '..':
                new_parts += [ os.path.pardir ]
            else:
                new_parts += [ part ]
         
        return os.path.join(*new_parts)


def get_all_files(path):
    file_names = []
    for listing in os.walk(path):
        file_names_in_dir = len(listing[2])
        if file_names_in_dir > 0: ## leaf level reached
            for base_name in listing[2]:
                file_name  = listing[0] + '/' + base_name
                file_names.append(file_name)
    return file_names


def get_all_dirs(path):
    dir_names = []
    for listing in os.walk(path):
        dir_names_in_dir = len(listing[1])
        if dir_names_in_dir > 0: ## leaf level reached
            for base_name in listing[1]:
                dir_name  = listing[0] + '/' + base_name
                dir_names.append(dir_name)
    return dir_names


def switch_suffix(file_name, new_suffix):
    return ".".join( file_name.split('.')[:-1] + [new_suffix])
    
def get_files_by_suffix(d, suffices):
    """Returns list of absolute pathes of the files in d with given suffix.
    If suffix is a list either suffix in the list is a match."""
    if not type(suffices) == list:
        suffices = [suffices]   
    dirs = get_all_dirs(d)
    out = []
    for d in dirs:
        for suffix in suffices:
            out += glob.glob(os.path.join(d, '*%s' % suffix))
    
    return out
    

def get_all_suffices(d):
    
    base_names = os.listdir(d)
    parts      = [ os.path.splitext(p) for p in base_names]  
    suffices   = [ suffix[1:] for (rest,suffix) in parts if rest and suffix]

    return sorted(list(set(suffices)))


def basename_minus_1(path):
    return os.path.splitext(os.path.basename(path))[0]


def homedir():
    return os.path.expanduser('~') + os.path.sep





if __name__ == "__main__":
    pass

    
    
    
    
    