"""This module contains a view classes for the presentation of v-cubes that should  
illustrate the concept of deployment of the meta information included in the v-cube. 
It is a frouzy draft build on questionable code fragments. Don't look at the code. 

The development of a better replacement module is not challenging in terms of algorithms. 
What it makes complex is that we want it to make good guesses on how 
the output is presented, given a query. This means that code has to cover many different 
situation.

TODO: needs clean-up
""" 


from numpy import squeeze, set_printoptions, prod

from . import vcubes 
from peewit.paths import PathKeeper

                  

from cut.baseclass  import Named
from cut.container import indices, multi_range
from cut.fileaccess import create_dir, save_file
from collections import OrderedDict

from copy import deepcopy
import os

set_printoptions(linewidth=160)

class Viewer(Named):
    """A Viewer shows results stored in v-cube. By now, it not interactive, that is
    all aggregations, selections, etc. have to be done on the v-cube first. """
    
    def init(self,
             vcube,
             mention_only = [],
             drob_header_1 = False,
             fixed_length=None):    
        pass
    
    ### output services textual

    def list_of_dims(self, vcube=None):
        
        if vcube is None:
            vcube = self.vcube 
        
        out = []
        
        lmax = max([ len(e) for e in vcube.dim_names()])
            
        for ename in vcube.dim_names():
            vcube_ = vcube.vcubes[ename]
            
            if self.mention_only == [] or ename in self.mention_only:
                if vcube.side_vcubes and ename in vcube.side_vcubes:
                    labels = [ vcube_.get_name() ]
                else:
                    labels = [ str(x) for x in vcube_.value_labels ]
                
                out += [ ename + ": " + ' ' * (lmax-len(ename)) + str(labels) ]
                
    
        return "\n".join(out)
    
    def print_vcube(self, vcube=None, squeeze_it=True, comment=None, title=None,
                          fixed_length=None):
        """Print the list of dimension including value labels and the cube itself. 
        Typically, there are many singleton dimensions. If squeeze_it, they are 
        removed for printing.
        """
        
        if not fixed_length and self.fixed_length:
            fixed_length = self.fixed_length
            
        if vcube is None:
            vcube = self.vcube 
        
        
        # is cube already objectized?
        if not ( vcube.cube.dtype == object and type(vcube.cube.ravel()[0]) == vcubes.CubeElement):
            cube = vcube.objectize_cube()
        else:
            cube = vcube.cube
        
        ## replace void entries
        bube_r  = vcube.bube.ravel()
        for i,e in  enumerate(cube.ravel()):
            if not bube_r[i]:
                e.value = '' 
                
        # set elements length for printing    
        if fixed_length:
            for e in cube.ravel():
                e.print_length_max = fixed_length                
        else:
            max_len = max([ len(str(e.value)) for e in cube.ravel() ] )
            for e in  cube.ravel():
                e.print_length_max = max_len
        
            
        self.send()
        
        
        if not self.drob_header_1:
            if title is None:
                title = '\nprint v-cube'
            self.send(title)
                
            self.send()
            self.send('v-cube: ', vcube.pathkeeper.basename(target=vcube.target_ename))
            if vcube.run_number:
                self.send('run-id: ', str(vcube.run_number))
            self.send('dims:   ', vcube.dim_names())
            self.send('target: ', vcube.target_ename)
            self.send()
        
        self.send( self.list_of_dims(vcube=vcube) )
        
        if not comment is None:
            self.send(comment,lb=2)

        if squeeze_it:
            cube  = squeeze(cube)
            dims  = [ ename for ename in vcube.dim_names()
                      if vcube.dim_sizes()[ename] > 1 ]  
                
            self.send('\nleft dims: ', dims, lb=2)
            self.send(cube)
        
        else:
            self.send(cube)
        
        self.send(lb=2)
        
            
    def print_with_marked_argaggregations(self, 
                                          aggregations, 
                                          squeeze_it=True,
                                          marker='*',
                                          fixed_length=8):
        """Prints cube where the results of arg-aggregations are 
        marked. For instance the maximizer along certain dimension are marked.
        If aggregations contain multiple arg-aggregations, the aggregations are
        jointly computed as done by VCube.argaggregation and thereof viewer makers 
        can be seen."""
        
        vcube = deepcopy(self.vcube)
        
        o_cube = vcube.objectize_cube( marker_default=' ' * len(marker))
        
        ic = vcube.argaggregations_indices(aggregations)  
          
        aggregants     = ic.value_names['aggregants']
        
        non_aggregants = [ ename for ename in vcube.dim_names() 
                           if ename not in aggregants ]
   
        na_sizes       = [ vcube.dim_sizes()[ename] for ename in non_aggregants  ]
        
        for mindex_fixed in multi_range(na_sizes):
            
            mindex_a = ic.cube[ mindex_fixed ]
            
            mindex   = [None] * (len(mindex_fixed) + len(mindex_a)) 
            
            for i, ename in enumerate(vcube.dim_names()):
                if ename in aggregants: 
                    j = aggregants.index(ename)
                    mindex[i] = mindex_a[j]
                else:
                    j = non_aggregants.index(ename)
                    mindex[i] = mindex_fixed[j]
            
            o_cube[tuple(mindex)].marker = marker 
                
        vcube.cube = o_cube

        agg_list = 'marked aggregation: ' + str(aggregations)
        
        self.print_vcube( vcube=vcube , squeeze_it= squeeze_it, comment=agg_list,
                          title='print v-cube with marked argagregations',
                          fixed_length=fixed_length)
        
        
    def print_argaggregations(self, aggregations, squeeze_it=True):
        """Prints the arguments that that are returned by
        arg-aggregation such as arg-max. 
        Output does not look nice. We have to learn how to control numpy output."""
        
        vcube = self.vcube
        
        ic = vcube.argaggregations_indices(aggregations)

        o_cube = ic.objectize_cube()
        
                
        aggregants     = ic.value_names['aggregants']
        
        non_aggregants = [ ename for ename in vcube.dim_names()
                           if ename not in aggregants ]
   
        na_sizes       = [ vcube.dim_sizes()[ename] for ename in non_aggregants  ]
        
        for mindex_fixed in multi_range(na_sizes):
            
            mindex_a = ic.cube[ mindex_fixed ]
            
            z        = list(zip(aggregants, mindex_a))
            
            values   = [ vcube.value_names[ename][i] for (ename,i) in z ]
        
            for i,value in enumerate(values):    
                o_cube[mindex_fixed][i].value = value
       
        
        ic.cube = o_cube
        agg_list = 'argaggregations: ' + str(aggregations)
        
        self.print_vcube( vcube=ic ,squeeze_it= squeeze_it,comment=agg_list,
                          title='print return values of arg-aggregations')
        
       
    lt_begin="""
\\begin{tabular}[t]"""

            
    lt_end="""
\\end{tabular}
"""


    def latex_table_snippet(self, row_dims,column_dims=[], source_info=True):
        
        assert 0, "table-snippets is currently out of service."
        
        #vcube = self.vcube.copy()
         
        non_singleton_dims = [ dim for dim, size in list(self.vcube.dim_sizes().items()) 
                               if size >  1 ]  
        
        
    
        assert row_dims
    
        if not column_dims:
            column_dims = [ename for ename in non_singleton_dims 
                           if not ename in row_dims ]
        
        
        for ename in row_dims + column_dims:
            self.vcube.value_names[ename] = [ self.tex_masking(vn) 
                                              for vn in self.vcube.value_names[ename] ]
        
        self.vcube.partially_permute(row_dims + column_dims)
        
        
        out  = self.lt_begin
        
        out += self.lt_layout_spec(row_dims, column_dims)

        out += self.lt_column_header(row_dims, column_dims) + "\\\\\n"

        out += self.lt_sub_table(row_dims, column_dims)

        if source_info:
            out += self.source_info_table_row(row_dims,column_dims)
            

        out += self.lt_end
        

        pk = self.vcube.pathkeeper
        snippet_dir = os.path.join(pk.dir_abs_pdir('snippets'),
                                    )
    
        path = os.path.join(snippet_dir, pk.basename() + '.tex')
        create_dir(snippet_dir)
    
        self.send('Save latex-snippet:\n', path)
        save_file(out,path)
        
    def source_info_table_row(self,row_dims,column_dims):
        dim_sizes = [ self.vcube.dim_sizes()[e] for e in column_dims ]
        n_total_column = len(row_dims) + prod(dim_sizes)
        out  = "\\\\ \n"
        out += "\multicolumn{" + str(n_total_column) + "}{r}{"
        out +=  self.source_info() + "}"
        return out
    
    def source_info(self):
        out  = "{\scriptsize {\\tt \hfill " 
        
        out += self.tex_masking( self.vcube.pathkeeper.basename() + " " + \
                                 str(self.vcube.run_number) )
        out += "}}"
        
        return out
    

    def tex_masking(self,s):
        return s.replace('_','\\_').replace('#','\\#').replace('-','{-}')

    def lt_column_header(self,row_dims, column_dims):
        dim_sizes = [ self.vcube.dim_sizes()[e] for e in column_dims ]
        out = ""
        for i,e in enumerate(column_dims):
            row = ""
            n_sub_columns = prod(dim_sizes[i+1:])
            
            for j,vn in enumerate(self.vcube.value_names[e]):
                row += "\\multicolumn{%d}{c|}{%s}" % (n_sub_columns,vn)
                row += "&" if not j == dim_sizes[i] -1 else  ""
            
            if i>0:
                
                row = "&".join(([ row ] * prod(dim_sizes[:i])))
                
            row += "\\\\ \\hline \n" if not i == len(column_dims)-1 else "" 
        
            out += row
        
            
        for r in row_dims:      
            r    
            out = self.lt_add_sv_column("", out)
        
        
        out += "\\\\ \\hline\n"

        return out

    def lt_sub_table(self,row_dims, column_dims, selected=OrderedDict()):
        
        head_dim   = row_dims[0 ]
        tail_dims  = row_dims[1:]
        vns        = self.vcube.value_names[head_dim]
        new_sub    = ""
        
        for i,vn in enumerate(vns):
            selected_           = deepcopy(selected)
            selected_[head_dim] = i
            #vns        = self.vcube.value_names[head_dim]
            if tail_dims:
                sub_t   = self.lt_sub_table(tail_dims, column_dims, selected_)
            else:
                sels    = [ ('seli', e_, k) for e_,k in selected_.items() ]
                sub     =  self.vcube.aggs(sels)
                    
                values  = sub.cube.ravel()
                balues  = sub.bube.ravel()
                    
                values_ = [ ( ("%.3f" % values[j])  if balues[j] else "" ) 
                                for j in indices(values) ]
                    
                sub_t = " & ".join(values_) 
            
            new_sub += self.lt_add_sv_column(vn, sub_t)

            if not i == len(vns)-1:
                new_sub += "\\\\ \n"  
            
                if sub_t.count("\n") > 1:
                    new_sub += "\\hline\n"
                
        
        return new_sub
         
            
    def lt_add_sv_column(self,value, sub_table):
        
        rows  = [ r.strip() for r in sub_table.split("\n") ]
        rows  = [ r for r in rows if r ]
        
        out = ""
        
        if rows:     
            rows_ = ["\\multirow{%d}{*}{%s}" % (len(rows),value) + "&" + rows[0]]        
            for row in rows[1:]:
                rows_ += ["&" + row ]
            
            out =  "\n".join(rows_)
            
            
        return out
        
    def lt_layout_spec(self,row_dims,colum_dims):
        out  = "{"
        out += "l|" * len(row_dims)
        n_col_total = prod([ self.vcube.dim_sizes()[e] for e in colum_dims ])
        out += "c|" * n_col_total
        out += "}\n\n"
        return out




        
    ### output services with plots
    
    def  build_plots(self, inner_dims=None, **plot_args):
        """TODO: major clean-up"""
        
        from pw_toolbox.plotting import MultiGraph 
        
        if not inner_dims:
            non_singletons = [ dim for dim, size in list(self.vcube.dim_sizes().items()) if size >  1 ]  
            inner_dims     = non_singletons[-2:]
            
        n_inner_dims = len(inner_dims)
 
        assert n_inner_dims in [ 1,2 ], inner_dims
        
        dummy_ag = [] ## list of dummy aggregations used to get vcube of desired shape 
        
        dim_x_axis = None
        dim_line_collors = None
        
        if n_inner_dims >= 1:
            dim_x_axis        = inner_dims[-1]
            
        if n_inner_dims >= 2:
            dim_line_collors = inner_dims[-2]
            
        assert dim_x_axis
        assert self.vcube.dim_sizes()[dim_x_axis] > 1, \
            "Two or more values for x-axis expected."
              
        vc_plot = self.vcube.copy()
        
        #print "AA0", vc_plot.target_ename, vc_plot.dim_sizes()
        AL = vcubes.Manipulation
        
        
        AL.make_singleton_parents(vc_plot, dim_x_axis)
        
        if dim_line_collors:
            AL.make_singleton_parents(vc_plot, dim_line_collors)
        
        vc_plot.check()
        
        AL.partially_permute(vc_plot, [dim_x_axis])
        vcube = self.vcube.copy()
        AL.partially_permute(vcube, [dim_x_axis])
        
        ## TODO: why did we do this?
        #vc_plot.target_ename = dim_x_axis
                
        for ename in vc_plot.dim_names():
            if vc_plot.dim_sizes()[ename] <= 1 and ename != dim_x_axis:
                AL.drop_singleton_parents(vc_plot, ename)
        
        vc_plot.cube = vc_plot.objectize_cube()
        

        x_label       = dim_x_axis
        try:
            x_values = [ float(x) for x in self.vcube.vcubes[dim_x_axis].value_labels ]
        except:
            x_values = indices( self.vcube.vcubes[dim_x_axis].value_labels)
            
        
        ## mid: multi-index-dict
        for mid in multi_range(list(vc_plot.dim_sizes().values()),vc_plot.dim_names()):
            
            mid.pop(dim_x_axis)
            
            vc_sub          = vcube.get_subcube(mid)
            
            ## y-label is value-label of target-dim
            
            #print(vc_sub)
            #print(vc_sub.value_labels)
            #print(vc_sub.dim_sizes())
            if len( set(vcube.value_labels) ) == 1:
                y_label = vc_sub.value_labels[0]
            else:
                y_label = vc_sub.target_ename
                
            vc_sub_shrinked = vc_sub.copy()
            for ename in vc_sub_shrinked.dim_names():
                if vc_sub_shrinked.dim_sizes()[ename] <= 1 and ename != dim_x_axis:
                    AL.drop_singleton_parents(vc_sub_shrinked, ename)
            
            if dim_line_collors:
                yshape   =  (vc_sub.dim_sizes()[dim_line_collors],
                             vc_sub.dim_sizes()[dim_x_axis] )
            else:
                yshape   =  (vc_sub.dim_sizes()[dim_x_axis] )
                
            x_value_labels = vc_sub.vcubes[dim_x_axis].value_labels
            
            #print(x_value_labels)
            if 'x_ticks' in plot_args:
                
                
                #print("FFF")
                #print(plot_args['x_ticks'])
                if all([ type(t) ==  int for t in plot_args['x_ticks']]):
                    ## x_ticks are taken as value-positions                    
                    for t in plot_args['x_ticks']:
                        assert t < len(x_value_labels), x_value_labels
                        pass
                    
                
                if all([ type(t) ==  type("") for t in plot_args['x_ticks']]):
                    for t in plot_args['x_ticks']:
                        assert t in x_value_labels
                    
                    plot_args['x_ticks'] = [ x_value_labels.index(t) 
                                             for t in plot_args['x_ticks']  ]                                 
                
                print(plot_args['x_ticks'])

                #plot_args['x_value_labels'] = [ x_value_labels[t] 
                #                                for t in plot_args['x_ticks']  ]
                
                
            y_matrix  = vc_sub_shrinked.cube.reshape(yshape)
            basename  = self.vcube.pathkeeper.basename()
            subtitle  = self.sub_subtitle(mid)
            title     = basename + '\n' + subtitle if subtitle else ''            
            
            if dim_line_collors:
                assert y_matrix.ndim == 2, vc_sub_shrinked.dim_names()
                AL.partially_permute(vc_sub, [ dim_line_collors, dim_x_axis])
                keys = vc_sub.vcubes[dim_line_collors].value_labels
                show_legend = True
                legend_title = dim_line_collors

            else:
                assert y_matrix.ndim == 1
                AL.partially_permute(vc_sub, [ dim_x_axis ] )
                keys = None
                show_legend = False
                legend_title = None
                
            
            args = deepcopy(plot_args)
            defaults = {     #'fignum': i,
                             'x_values' : x_values,
                             'y_matrix' : y_matrix,
                             'keys'     : keys,
                             'x_label'  : x_label,
                             'y_label'  : y_label,
                             'show_legend'  : show_legend,
                             'title'        : title,
                             'legend_title' : legend_title,
                             'show_legend_title': 1,
                             #'x_margin'     : 0.01
                             'source_id': self.vcube.run_number,
                             'x_value_labels' : x_value_labels,
                             'x_ticks_all_values': True,
                        }
            
            for key in defaults.keys():
                if key not in args:                    
                    args[key] = defaults[key]
                        
            p = MultiGraph(**args)
            p.pre()
            p.plot_content()
            p.plot_rest()
            if subtitle:
                filename = basename + '--' + subtitle
            else:
                filename = basename
                
            p.file_path = os.path.join(self.ab_plotdir(), filename )
            
            vc_plot.set_value(mid, p)
            
        self.plot_cube = vc_plot
              
            
    def sub_subtitle(self, mid):
        
        out = ""
        for ename,i in mid.items():
            
            if self.vcube.dim_sizes()[ename] > 1:
                if out:
                    out += '___'
                out += ename + '__' + self.vcube.vcubes[ename].value_labels[i]
        
        return out
    
    def save_plots(self,plot_cube=None, format='pdf'):
        
        import pylab 
        
        plot_cube =  plot_cube if plot_cube  else self.plot_cube
        
        plots = list(plot_cube.cube.ravel()) 
            
        for p in plots:
            path        = (p.file_path + '.' + format)
            ab_plot_dir = os.path.dirname(path)
            
            self.send('Save plot:\n%s' % path)
            create_dir(ab_plot_dir)
            
            if not PathKeeper.use_pylab:
                self.bugger('Please install matplotlib and set PathKeeper.use_pylab true')
            
            
            pylab.figure(p.fignum)
            pylab.savefig(path, dpi=300, format=format)

        
    def ab_plotdir(self):
        return os.path.join( self.vcube.pathkeeper.dir_abs_pdir('plot'), self.vcube.pathkeeper.basename() )



            


if __name__ == '__main__':

    pass
    












    