from cut.baseclass import Named
from cut.container import  indices, transpose_list,has_unique_value

from numpy import mean, array, random, histogram, arange, nanmax,nanmin,dot, reshape,\
                  sin ,nanargmax,nanargmin,log
from copy import  deepcopy

from cut.numpytools import  select_rows, a2sign, lunion, number_like
try:
    import matplotlib.pyplot as pyplot
except: 
    pass
from numpy import linalg
from abc import ABCMeta, abstractmethod


class MultiPlot(Named, metaclass=ABCMeta):
    """A multi-plot is a 2d plot where a third dimension is 
    encoded by colors (or textures). There are the dimensions 
    x-axis, y-axis, and z-axis where the latter is the color
    dimension.
    This class is a basis sauce for different multi plot classes
    and cannot be used itself. Most of its members are concerned
    with the meta-data and layout of the plot.
    """
       
    def init(self,
                fignum    = None,
                x_matrix     = None, # for each part i there are list of corresponding   
                x_values     = None, # In some cases the parts share the x-values
                y_matrix     = None, # x- and y-values;
                                   # the way the data is read is left to the descendants.
                keys         = None, # names of the parts
                colors       = ['b','g','y','m','r','c'] * 10, # the colors used for the 
                
                x_values_all = None, # union of values in x_matrix
                y_values_all = None, # ..
            
                x_lim            = None, # left and right border of the plot 
                x_margin         = None,  # set x_lim relative the min/max of x_values_all
                y_lim            = None, # top and bottum border of the plot
                y_margin         = None,  # ..
                                    
                
                x_ticks_all_values = False, # draw x-lines, e.g. shattered lines parallel 
                                            # to y-axis for any x-value
                x_ticks            = None,  # draw x-lines for a values in this list
                y_ticks            = None,  # ..
                y_ticks_indices    = [],    
                y_ticks_additional = [], 
                tick_max           = False, # draw y-lines at greatest y-value for each part
                tick_argmax        = False, # draw x-line at the maximizing x-value
                tick_min           = False, # ..
                tick_argmin        = False, # ..
                show_grid          = True,  # switch for plotting x/y-lines
                
                               
                title           = "Multi Plot",
                x_label         = "x",
                y_label         = "y",
                text            = None,
                legend_mapping  = lambda x:str(x),
                x_value_labels = None,
                show_legend     = True,
                legend_title    = None,
                show_legend_title=False,
                source_id        = None,
                ):
        
        self.x_min = None
        self.x_max = None
        self.y_min = None
        self.y_max = None
        
        pass
    
    def expand_x_values_to_matrix(self):    
        self.x_matrix = [ None ] * self.n_parts
        for i in range(self.n_parts):
            self.x_matrix[i] = self.x_values
        
    def set_extrema(self):
        
        self.n_parts = len(self.y_matrix)
        
        # x-values
        if self.x_matrix != None:
            self.x_values_all = []
            for values in self.x_matrix:
                #print values
                self.x_values_all = lunion(self.x_values_all, values)
        elif self.x_values != None:
            self.x_values_all = self.x_values
        else:
            self.bugger()

        if self.x_values_all and number_like(self.x_values_all[0]):
            
            self.x_min = nanmin(self.x_values_all)
            self.x_max = nanmax(self.x_values_all)
        

        # y-values
        self.y_values_all = []
        for values in self.y_matrix:
            values_ = values.ravel()
            self.y_values_all = lunion(self.y_values_all, values_)
        
        self.y_min = nanmin(self.y_values_all)
        self.y_max = nanmax(self.y_values_all)

        self.y_maxima    = [ None ] * self.n_parts
        self.y_argmaxima = [ None ] * self.n_parts
        
        self.y_minima    = [ None ] * self.n_parts
        self.y_argminima = [ None ] * self.n_parts
        
        for i in indices(self.y_matrix):
            #print i, self.n_parts
            self.y_maxima[i]    = nanmax(self.y_matrix[i])
            self.y_argmaxima[i] = self.x_matrix[i][ nanargmax(self.y_matrix[i])] 
        
            self.y_minima[i]    = nanmin(self.y_matrix[i])
            self.y_argminima[i] = self.x_matrix[i][ nanargmin(self.y_matrix[i])] 
        
        
       
    
 
    def set_ticks(self):
        ## ticks
        """
                x_ticks_all_values        = True,
                y_ticks            = None,
                y_ticks_indices    = [],
                y_ticks_additional = [], 
                tick_max           = False,
                tick_min           = False,
                show_grid          = True,"""
                
        ## x-axis
        
        
        self.x_ticks_to_be_added = []
        
        if self.tick_argmax:
            self.x_ticks_to_be_added.extend(self.y_argmaxima)
        
        if self.tick_argmin:
            self.x_ticks_to_be_added.extend(self.y_argminima)
        
        
        if self.x_ticks_all_values:
            self.x_ticks_to_be_added.extend( self.x_values_all )
        
    
        ## y-axis
        ## To obtain no ticks, use y_ticks=[],
        ## y_ticks=None results in default ticks. 
        ## But only, if there are other y_ticks inputs
        
        
        if self.y_ticks == None:
            self.y_ticks
            y_ticks_was_none = True
            # keep the None and give it a try          
            self.y_ticks = []
        else:
            y_ticks_was_none = False
        
        
        self.y_ticks_to_be_added = []
        
    
        if self.tick_max:
            self.y_ticks_to_be_added.append(self.y_max)
        if self.tick_max:
            self.y_ticks_to_be_added.extend(self.y_maxima)
                
            
        if self.tick_min:
            self.y_ticks_to_be_added.append(self.y_min)
        if self.tick_min:
            self.y_ticks_to_be_added.extend(self.y_minima)
        

        self.y_ticks_to_be_added.extend(self.y_ticks_additional) 
    
        for part, x_index in self.y_ticks_indices:
            self.y_ticks_to_be_added.append(self.y_values[part][x_index])
        
        
        # still empty? Recover None  
        if self.y_ticks == [] and y_ticks_was_none: 
            self.y_ticks = None
            
        
                         

    def set_lables(self):
        pass    

            
    def set_lims(self):
        """sets the coordinates of the frame."""
        
        if self.x_lim == None and self.x_min and self.x_max:
            
            total_width = self.x_max - self.x_min
            assert total_width >= 0
            if not self.x_margin is None:
                assert self.x_margin >= 0
            
                self.x_lim = [ self.x_min  -  self.x_margin * total_width, 
                               self.x_max  +  self.x_margin * total_width  ]
        
        if self.y_lim == None and self.y_min and self.y_max:
            
            total_heigth = self.y_max - self.y_min
            assert total_heigth >= 0
            
            if not self.y_margin is None:
                assert self.y_margin >= 0
            
                if self.y_min > 0:
                    self.y_lim = [ self.y_min  -  self.y_margin * total_heigth/2.,
                                   self.y_max  +  self.y_margin * total_heigth  ]
                else:
                    self.y_lim = [ self.y_min  -  self.y_margin * total_heigth /2., 
                                   self.y_max  +  self.y_margin * total_heigth  ]
       
        
    @abstractmethod
    def plot_content(self):
        """Send the commands that plot the content, e.g. values."""
        pass
    


    def plot_rest(self):
        """Send the command that set the extras."""
         
        if not self.x_lim is None:
            
            pyplot.xlim( self.x_lim[0], self.x_lim[1] )
            
        if not self.y_lim is None:
            pyplot.ylim(self.y_lim[0], self.y_lim[1])
        
        
        ## x_ticks
        if self.x_ticks is None:
            if self.x_ticks_all_values:
                self.x_ticks = list(self.x_values)
            else:
                ## use pyplots ticks
                self.x_ticks = list(pyplot.xticks()[0])
        
        
        self.x_ticks.extend(self.x_ticks_to_be_added)
        assert not self.x_ticks is None
        
        
        if self.x_value_labels is None:
            if all([ x == int(x) for x in self.x_ticks]):
                self.x_value_labels = [ "%d" % int(x) for x in self.x_ticks ]
            else:
                self.x_value_labels = [  "%0.2f" % x  for x in self.x_ticks ]
        
        else:
            assert len(self.x_value_labels) is len(self.x_values)
            
        if self.x_value_labels is None:
            pyplot.xticks( self.x_ticks )
        else:
            pyplot.xticks( self.x_ticks,  self.x_value_labels)  
    
        ## y_ticks
        if self.y_ticks is None:
            self.y_ticks = list( pyplot.yticks()[0] )
            
        assert type(self.y_ticks) is list
        
        self.y_ticks.extend(self.y_ticks_to_be_added)
    
    
        pyplot.yticks(self.y_ticks, [ '%1.2f'  % y for y in self.y_ticks])
    
    
        pyplot.grid(self.show_grid)
        
        pyplot.title(self.title)
        pyplot.ylabel(self.y_label)
        pyplot.xlabel(self.x_label)
        if self.show_legend:
            
            leg= pyplot.legend( [ self.graphs[i][0]  for i in range(self.n_parts)], 
                          [ self.legend_mapping(x) for x in self.keys], 
                          loc = 'best',
                          shadow=True,
                          #fancybox=True
                        )
            
            if self.show_legend_title:
                leg.set_title(self.legend_title) 
        
        if self.source_id:
            
            xr = pyplot.xlim()[1]
            yb = pyplot.ylim()[0]
            xd = pyplot.xlim()[1] - pyplot.xlim()[0]
            yd = pyplot.ylim()[1] - pyplot.ylim()[0]
            

            pyplot.text(xr - 0.09 * xd,
                       yb - 0.07 * yd,
                       str(self.source_id), 
                       {'fontsize':6}
                       )
            pyplot.text
          

    def pre(self):
        #id = 1 if self.fignum is None else self.fignum
        
        self.fignum = pyplot.figure().number
        self.set_x_values()
        self.set_y_values()
        self.set_extrema()
        self.set_ticks()
        self.set_lables()
        self.set_lims()
        
        
    def go(self):
        self.pre()
        self.plot_content()
        self.plot_rest()
        pyplot.show()

    def show(self):
        pyplot.show()

    def scalar_color(self,a):
        """experimental .."""
        b = 1.-a
        A = [  
              [2,0,1],
              [0,2,3],
              [4,2,1]
            ]
        A = array(A, dtype=float)
        
        x = reshape(array((b,b,a)), (3,1))
        y = dot(A,x)
        y = y /y.sum()
        y = y[[0,1,2]]
        return tuple(y.flatten())

    def add_horizontal_line(self,plot,value,ls='-', lw=0.2, color='black'):
        pyplot.figure(plot.fignum, )
        pyplot.plot(plot.x_lim, [value] * 2, ls=ls, lw=lw, color=color)


    def add_vertical_line(self,plot,value,ls='-', lw=0.2, color='black'):
        pyplot.figure(plot.fignum)
        
        pyplot.plot([value]*2, plot.y_lim, ls=ls, lw=lw, color=color)



    def add_marking_x(self,plot,value,ls='-', lw=1.0, color='blue',size=0.04):
        pyplot.figure(plot.fignum)
        y_low, y_top = self.y_lim
        pyplot.plot([value]*2, [y_low, y_low + (y_top -y_low)*size], 
                    ls=ls, lw=lw, color=color)


    def add_marking_y(self,plot,value,ls='-', lw=1.0, color='blue',size=0.04):
        pyplot.figure(plot.fignum)
        x_low, x_top = self.x_lim
        pyplot.plot([x_low, x_low + (x_top -x_low)*size], [value]*2,  
                    ls=ls, lw=lw, color=color)


   

class MultiGraph(MultiPlot):
    """Plots multiple graphs with a common set of x-values. The y-values are given 
    as array where each row represent a part."""
    
    def init(self,  
                y_matrix        = None,
                x_values        = None, 
                x_joint         = None,  # shorter name for x_values_joint
                x_row           = None,  # index of y_matrix that is taken as x_values_joint
                
                
                scaled          = False, # scalling of the y_values that at index i_s 
                                         # the y_value is 1 for all parts 
                rel_scale_point = 0.2,   # chose i_s relative to number of x_values 

                marker          = 's',  
                markersize      = 4,
                
                    
               ):
        pass
    
    def set_x_values(self):
        """sets x_values and x_inidices"""
        
        self.n_parts = len(self.y_matrix)
        
        if self.x_values != None:
            pass
        
        elif self.x_row != None:
            assert type(self.x_row) == type(2) 
            self.x_values = self.y_matrix[self.x_row]
        else:
            lengths = [ len(row) for row in self.y_matrix ]
            assert has_unique_value(lengths)
            self.x_values = indices(self.y_matrix[0])
        
        self.expand_x_values_to_matrix()
 
    
    def set_y_values(self):

        if self.scaled:
            scale_point   = int( self.x_values_joint * self.rel_scale_point )
            self.y_values = [ v / v[scale_point]  for v in self.y_matrix]
        else:
            self.y_values = self.y_matrix
            
        
    def plot_content(self):

        self.graphs = [ None ] * self.n_parts

        for i in range(self.n_parts):
            
            self.graphs[i] = pyplot.plot( self.x_values, 
                        self.y_values[i], 
                        '-',
                        markersize=self.markersize,
                        marker=self.marker, 
                        color=self.colors[i],
                        antialiased = True,
                      )
            
            
 
class FunctionPlot(MultiPlot):
    
    def init(self,
                # does not work: sin cannot be deepcopied
                #functions  = [FunctionWrapper(sin)],
                functions  = [],
                n_points   = 200,
                x_interval = [-4,4],  # x_intervall = [x_min, x_max]
                x_margin   =  0,
                x_ticks=[0], 
                y_ticks=[0],
                tick_max    = True,
                tick_argmax = True,
                linewidth   = 1,
                ):
        pass
    
        for i in indices(self.functions):
            try:
                #print self.functions[i]
                self.functions[i] = self.functions[i].apply
            except:
                pass
                #print "nop"
        
        #self.functions = None    
        #print self.functions
        
        if not self.keys:
            try:        
                self.keys = [ f.__name__ for f in self.functions ]
            except: 
                self.keys = indices(self.functions)
                
    def set_x_values(self):
        """
        sets x_values and x_inidices 
        """
        if not self.x_values:
            assert self.n_points >= 2
            interval_size = self.x_interval[1] - self.x_interval[0] + 0.
            assert interval_size > 0
            step_size = interval_size/(self.n_points-1)
            self.x_values = arange(self.x_interval[0], self.x_interval[1]+step_size, step_size)
            #print self.x_values

        self.x_matrix = []
        for  i in indices(self.functions):
            self.x_matrix.append(deepcopy(self.x_values))


        
    def set_y_values(self):
        
        self.n_parts  = len(self.functions)
        self.y_matrix = [ None ] * self.n_parts
        print(self.functions)
        for i in indices(self.functions):
            
            self.y_matrix[i] = [ self.functions[i](x) for x in self.x_values ]
    
        self.y_matrix = array(self.y_matrix)
         
    def plot_content(self):

        self.graphs = [ None ] * self.n_parts

        for i in indices(self.functions):
            self.graphs[i] = pyplot.plot( self.x_values, 
                                        self.y_matrix[i], 
                                        '-', 
                                        color=self.colors[i],
                                        linewidth=self.linewidth,
                                      )


class DataPlot(MultiPlot):
    """calls for plotting 2-dimnesional labled data set 
    along with seperating hyperplanes (lines)""" 
    
    def init(self,
                data_matrix   = None, # input matrix, only tow columns are to plotted
                data_text     = None, 
                output_vector = None, # number of distinct values is the number of parts
                lines         = [],
                line_colors   = ['b','g','r'],
                line_text     = [],
                dim_selection = [0,1], # what dims of X are shown
                marker        = '.',
                colors             = ['r','g','m','y','b','c'] * 10 ,
                show_grid=False,
                #legend_mapping  = lambda x: '%+.0f' % x ,
                ):
        
        self.line_colors = self.line_colors * 2000
        pass
    
    
    def build_class_parts(self):
        
        classes = sorted(list(set( list(self.output_vector.ravel()))))
        self.keys = classes
        
        self.n_parts = len(classes)
        
        parts = []
        
        for i in indices(classes):
            data_part = select_rows(self.data_matrix, self.output_vector == classes[i])
            x_part = data_part[:,self.dim_selection[0]].ravel()
            y_part = data_part[:,self.dim_selection[1]].ravel()
            parts.append( (x_part, y_part) )
    
        self.parts =parts
    
    def set_x_values(self):
        
        self.build_class_parts()
        self.n_parts = len(self.parts)
        
        self.x_matrix = [] 
        
        for part in self.parts:
            self.x_matrix.append(list(part[0]) )
   

    def set_y_values(self):
        self.y_matrix = []
        for part in self.parts:
            self.y_matrix.append(part[1])
            

    def plot_content(self):

        self.graphs = [ None ] * self.n_parts

        for i in range(self.n_parts):
            
            x_part, y_part = self.parts[i]
            self.graphs[i]  = pyplot.plot( x_part,
                                         y_part, 
                                         '',
                                         marker=self.marker, 
                                         color=self.colors[i],
                                         linewidth=0,
                                       )
            if i > 0:
                break
        
        
        Z = self.data_matrix
        dd = 0.02
        if self.data_text != None:
            for i in indices(Z):
                pyplot.text(Z[i][0]+dd,Z[i][1]+dd, self.data_text[i],
                           fontsize=9,
                           horizontalalignment='left',
                           verticalalignment='bottom')
        
        self.plot_lines( self.lines )
        
 
    def plot_lines(self, lines=None):
        
        assert self.x_min
        
        if not lines:
            lines = self.lines
        
        
        
        colors = self.line_colors  
        
        self.plots2 = [ None ] * len(self.lines)
        for i in indices(lines):
            line = lines[i]
            
            v_0, v_1 = line
            
            v_1  = v_1/linalg.norm(v_1)
            
            def h(r):
                return v_0 + r * v_1
            
            
            P = list(arange(self.x_min * 100 , self.x_max * 100 , 1)/100.)
            P = list(arange(-10 , 10, 1))
            
            #P = list(arange(-5 * 100 , 10 * 100 , 1)/100.)
            
            R = [ h(r) for r in P ]
            
            x,y = transpose_list(R)
            
            #x_part , y_part = transpose_list(R)
            
            
            self.plots2[i]  = pyplot.plot( x,
                                          y,  
                                          color = colors[i],
                                          linewidth = 0.7,
                                       )
            
            if self.line_text:
                md = 0.1
                v_3 = v_0 + v_1 * (0.8 + md * i)
                pyplot.text(v_3[0],v_3[1], self.line_text[i],color=colors[i],fontsize=10)
        
        
        
    def gos(self):
        
        self.set_x_values()
        self.set_y_values()
        self.set_extrema()
        self.set_ticks()
        self.set_lims()
        self.plot_content()
        self.plot_rest()
        pyplot.show()

    
class Histogram(MultiPlot):
    
    def init(self,
                n_bins        =  10, 
                bin_per_value = False,
                title         = "Histogramm",
                x_label       = "ranges of values",
                y_label       = "frequencies",
                plot_mean_bar = True,
                normed        = True,
                relative_bar_width             = 0.84,
                adjust_barwidth_to_cardinality = True,
                x_slice       = None,
                rel_y_lim_factor = 1.1,
                show_legend     = False
                ):
        #broken_bugger # class has not be rewritten by last module rewrite
        pass
    
    def set_x_values(self):
        pass
    
    def set_y_values(self):
        self.n_parts      = len(self.y_matrix) 
        self.y_raw_values = self.y_matrix
        
        self.y_raw_values_all = list(set(pyplot.flatten(self.y_raw_values))) 
        
        self.y_raw_min = nanmin(self.y_raw_values_all)
        self.y_raw_max = nanmax(self.y_raw_values_all)
        
        if not self.x_slice:
            self.x_slice = [self.y_raw_min, self.y_raw_max]

        if self.bin_per_value:
            
            assert all( [ int(x) == x for x in self.y_raw_values_all ] )
            
            self.delta = 1.
            self.total_bar_width = self.delta * self.relative_bar_width

            rel_bar_position_0 = - self.total_bar_width/2.
            #print "RR" ,rel_bar_position_0
            self.n_bins = arange(self.x_slice[0], self.x_slice[1])
        
        else:
            
            self.delta = (self.x_slice[1] - self.x_slice[0]) / float(self.n_bins)
            self.total_bar_width = self.delta * self.relative_bar_width
            rel_bar_position_0 = (self.delta - self.total_bar_width)/2
        
        
                    
        self.y_matrix      = [None] * self.n_parts
        self.bar_positions = [None] * self.n_parts
        self.bar_widths    = [None] * self.n_parts
        self.n_values      = [len(values) for values in self.y_raw_values]
        self.n_all_values  = sum(self.n_values)
        assert self.n_parts > 0
        
        self.x_matrix = [ None ] * self.n_parts
        
        for i in  range(self.n_parts):
            
            freq, bins = \
                histogram(self.y_raw_values[i], bins=self.n_bins, range=self.x_slice)
            
            self.y_matrix[i] = freq
            self.x_matrix[i] = bins[:-1]
            
            print(self.y_matrix[i], self.x_matrix[i]) 
            if self.adjust_barwidth_to_cardinality:
                bar_width_fraction  =  self.n_values[i] / float(self.n_all_values)
            else:
                bar_width_fraction  =  1. / self.n_parts
            
            self.bar_widths[i] =  self.total_bar_width * bar_width_fraction
            
            if i == 0:
                self.bar_positions[i] = [ x + rel_bar_position_0 for x in self.x_matrix[i]] 
                #print "TTT" , self.bar_positions
            else:
                self.bar_positions[i] = [ x +  self.bar_widths[i-1] 
                                          for x in self.bar_positions[i-1]]
            if self.normed:
                self.y_matrix[i] = self.y_matrix[i] / float(sum(self.y_matrix[i]))
            
     
       
    def set_ticks(self):
        
        """Build the ticks for x and y axis."""
        MultiPlot.set_ticks(self)
        
        return 
    
    def set_lables(self):
        MultiPlot.set_lables(self)
        
        if self.normed:     
            self.y_label = "relative frequencies"
        else:
            self.y_label = "absolute frequencies"
 
    def set_lims(self):
        
        MultiPlot.set_lims(self)
        
        if self.normed:
            
            if self.y_max < 0.6:
                self.y_lim =  [ 0, self.y_max * self.rel_y_lim_factor]
            else:
                self.y_lim =  [ 0,  1. * self.rel_y_lim_factor ] 
            
            self.y_label = "relative frequencies"
            
         
        else:
            self.y_lim = [ 0, self.y_max * self.rel_y_lim_factor]
            self.y_label = "absolute frequencies"
            if not self.y_ticks_absolute:
                self.y_ticks_absolute = []
        
            
        
    def plot_content(self):
        
        self.plots     = [ None ] * self.n_parts
        self.mean_bars = [ None ] * self.n_parts
        
        for i in range(self.n_parts):
    
            print("HHH", self.y_matrix[i])
            print("HHH", self.bar_positions[i])
            print(len(self.bar_positions[i]), len(self.y_matrix[i]))
            
            self.plots[i]  = pyplot.bar(self.bar_positions[i],
                                       self.y_matrix[i], 
                                       self.bar_widths[i], 
                                       color=self.colors[i] )         
                                        
            # problem need y_lim
            if 0 and self.plot_mean_bar:
                self.mean_bars = pyplot.bar([ mean(self.y_matrix[i]) ] , 
                                           [self.y_lim[1]], 
                                           self.delta * 0.1, 
                                           color=self.colors[i])
        



if __name__ == '__main__':

    if 0:
        d = 30
        n = 3
        X = (random.rand(n,d)-0.5) * 2
        p = Histogram(y_matrix=X, keys=[1,2,3], show_legend=True)
        p.go()
        
    if 0:
        d = 10
        n = 60
        X = (random.rand(n,d)-0.5) * 2
        y = a2sign(random.rand(n,1) - 0.5 )
        #y = col(ones(n))
        #y = col(array([-1,1]))
        
        dp = DataPlot(
                      data_matrix   = X,
                      output_vector = y,
                      lines      = [],
                    )
        dp.go()
    
    if 0:  
        def g(x):
            return  - x * log( x-0.1)
        
        P = FunctionPlot(functions=[  g, sin ], 
                         x_interval = [0,1], 
                         x_margin   =  0,
                         x_ticks=[0], 
                         y_ticks=[0],
                         tick_max    = True,
                         tick_argmax = True,
                         #x_values = [1,2,3,4],
                         #x_ticks_all_values=True,
                         )
        P.go()
         
         
    if 0:
        x = arange(0,1.1,0.1)
        Y = random.rand(3,11)
        p = MultiGraph(x_values  = x,
                      y_matrix = Y,
                      keys=['a','b','c'])
        p.go()
       
       
            
    
    pyplot.show()
