import unittest 

from peewit.core   import Experiment, ENode
from peewit.vcubes import VCube, Manipulation
from peewit.presentation import Viewer
from collections import OrderedDict
from cut.baseclass import Named

from numpy import array
import project_conf
import main

#reload(project_conf)

PK = project_conf.PK
VCube.pathkeeper = PK

class base(ENode):
    def descent(self):
        return [ 100,200,300,400 ]
 
class factor(ENode):
    def descent(self):
        
        
        return( 2,1,3)
        return a
        
    def ascent(self,first):
        self.vc_first = first
        
        #first.show()
        r = first.cube.size
        return [ r ] 
        
class another_factor(ENode):
    def descent(self):
        A = Named()
        A.coin = 3
        A.name = "small"
        self.A = A
        B = Named()
        B.name  = "medium"
        B.coin = 5
        
        return [ A, B ]

class offset(ENode):
    def descent(self, first):   
        self.first_dvc = first
        
        return [ 20,40 ]
    
class first(ENode):
    def descent(self, base, factor, another_factor):
        a = base + factor*another_factor.coin 
        return [ a  ]

class second(ENode):
    
    def descent(self, base, factor, offset):
        
        b = base +  factor.get_value() + offset
        return [ b ]
 
etree = [ base(),
          [ [ factor(), 
              another_factor(),
              first(), 
            ],
            [ offset(), 
              second(),
            ] 
           ]
        ]

## TestCase: a bundle of tests
class Branching(unittest.TestCase):
    
    def setUp(self):
        
        E = Experiment( pathkeeper = PK,
                        version    = 0
                      )
        self.E  = E
        self.PK = PK
        self.etree = etree
        E.define(self.etree)
        Experiment.treenames = set()
        self.vc = vc = E.run()
        
        
        
        for e in E.tree_runner.enodes:
            self.__dict__[e.ename] = e
    
        vc.check()
        vc.save()
        self.basename = vc.pathkeeper.basename()
        
    
    def test_branch_pvc(self):
        
        vc1 = VCube(basename=self.basename)
        vc1.show()
         
        etree = [ base(),
          [ [ factor(), 
              another_factor(),
              first(), 
            ],
            [ offset(), 
              second(pdvc=vc1),
            ] 
           ]
        ]
        E = Experiment( pathkeeper = PK,
                        version    = 1
                      )
        
        E.define(self.etree)
        Experiment.treenames = set()
        
        vc2 = E.run()
        
        vc2.show()
        
    
    def test_ascent_arg(self): 
        
        factor = self.factor
        
        factor.avc.show()
         
        self.assert_(factor.input_nodes_post == set([self.first]))
        print(factor.dependencies_ascent)
        print(factor.avc.side_vcubes)
        self.assert_(factor.dependencies_ascent == [ self.base,
                                                     self.first,
                                                   ] 
                     ) 
        
        vc     = factor.vc_first
           
        dim_names = ['factor','another_factor', 'first']
        
        self.assert_(vc.dim_names() == dim_names)
        self.assert_(list(vc.dim_sizes().values()) == [3, 2, 1])
        
        
        m_index = dict(list(zip(dim_names,[2,1,0])))
        cube_value = vc.get_value(m_index)
        #print(cube_value)
        self.assert_(cube_value == 415)
        
        vc.show()
        vc.check()
        
        factor.avc.drop('first')
            
        factor.avc.show()
        
    
    def test_enode_refs(self):
        offset = self.E.get_node('offset')
        self.assert_(offset is self.offset)
        self.assert_(self.E.get_node(['base','another_factor']) is self.another_factor)
     
    def test_arg_ag(self): 
 
        vc     = self.factor.vc_first
 
        
        
        arg1,arg2 = vc.arg_ag(['argmin', 'factor','another_factor'])
        
        self.assert_(not arg1 is self.another_factor.A)
        self.assert_(arg2 is self.another_factor.A)
    
    #@unittest.skip('') 
    def test_ag(self): 
 
        vc   =  self.factor.vc_first
        vc_  =  vc.aggs([('max', 'factor', 'another_factor')])
        
        vi = Viewer(vc_)
        vi.print_vcube()
        

    def test_s_ancestor_access(self):

#         D= {'has_shadow_values': False, 
#             'name': None, 
#             'basename': None, 
#             'read_failed': False, 
#             'bube': array([ True,  True], dtype=bool), 
#             'run_number': None, 
#             'vcubes': OrderedDict([('another_factor', <peewit.vcubes.VCube object at 0xad4484c>)]), 
#             'sube': array([], dtype=object), 
#             'verbosity': 1, 
#             'target_ename': 'another_factor', 
#             'value_labels': ['small', 'medium']}
           
        vc = self.offset.first_dvc
        vc.check()
        self.assert_(type(vc) is VCube)
        
        self.assert_(vc.dim_names() == ['factor', 'another_factor', 'first'])
        

if __name__ == "__main__":
    
    main.run_name('ascent.Branching.test_branch_pvc')

        
        