import unittest 

from peewit.core   import Experiment, ENode
from peewit.vcubes import VCube
from peewit.presentation import Viewer
from collections import OrderedDict
from cut.baseclass import Named

from project_conf   import PK

class Tube(Named):
    def init(self, line=None):
        self.name = self.classname() + '_' + str(line)

class rabbit(ENode):
    def descent(self):
        return [ 'a', 'b' , 'c', 'd']
 
class carrot(ENode):
    def descent(self):
        A = Tube(line=1)
        B = Tube(line=2)
        C = Tube(line=3)        
        return [ A, B, C ]

class kasha(ENode):
    def descent(self, rabbit, carrot ):
        
        def h(x,y,z):
            return  abs(("abcd".index(x)**2 - 5*z  ) * y.line )
            
        d2 = h(rabbit, carrot, 2)
        d3 = h(rabbit, carrot, 3)
        
        return [ d2, d3 ]
     
    def explicit_descent_labels(self):
        return ["He", "Hu"]


class mana(kasha):
    
    def descent(self, rabbit, carrot, mana):
        
        x,y = kasha.descent(self,rabbit,carrot)
        #print x,y, mana['carrot'], mana['rabbit']
        x += 10*mana.carrot[0]
        y += 10*mana.rabbit[1]
    
        return [x,y]
    
    def initial_siblings(self, carrot):
        
        box = Named()
        box.rabbit = [7,7 + carrot.line]
        box.carrot = [-8,-10]
        
        return box



etree = [ rabbit(),
          carrot(), 
          kasha(), 
        ]

## TestCase: a bundle of tests
class LinearTree(unittest.TestCase):
    
    ## 
    def setUp(self):
        

        E = Experiment( pathkeeper = PK,
                        version    = 0
                      )
        
        
        self.E  = E
        self.PK = PK
        self.etree = etree
        E.define(self.etree)
        ## need to overwrite incremental count in treename
        ## since setUp is called before each test
        Experiment.treenames = set()
        
    ## each function that have a name with prefix "test" 
    ## is a test
    def test_conf(self): 
        
        self.assert_(self.PK.basename().startswith('basics') )

    def test_define(self): 
        
        
        self.assert_( etree[0] is etree[2].input_node_map[etree[0].ename])  
        
        
    def test_run(self): 
        vc = self.E.run()
        vc.show()
    
    def test_svalues(self): 
        etree_ = etree[:]
        etree_[-1] = mana()
        
        
        E = Experiment( pathkeeper = PK,
                        version    = 1
                      )
        
        E.define(etree_)
        vc = E.run()
        vi = Viewer(vc)   
        vi.print_vcube(squeeze_it=1) 
        
        
        
        
if __name__ == "__main__":
    import main
    main.run_name('basics.LinearTree.test_run')
    
       
        
    
    