"""
This is the main module for testing the peewit code by means of units tests.
For those how are not familiar with python/java unit tests we have added explanations
to test module 'pre_run' (called by this module) that help you to separate peewit 
elements from unit-test elements. This is important since we consider the tests as 
part or at least as supplement of the documentations.

The notion 'unit tests' refers to testing of well-rounded sub-unit of
a software system, opposed to 'integration tests' that test checks wether 
the compounds of the system work together correctly. 

For python, java and presumeably for many other languages there are frameworks
to support unit testing. To us the two main features of unit test framework are

 1) distinction failure vs error: the former is a negative test result and latter 
    means no test result. Both are both valuable as both point you to a problem 
    in the system code and/or in the test code. But if it is a failure, then it it 
    unlikely to be a 'simple bug' in the test code.  

 2) summarization: even if some test break it goes through all the test and reports. 
    This way you get a better view on the situation than by solving one indicated issue 
    after another without seeing what the subsequent failures are.

For different reasons we us the out-dated test framework ('unittest' module at python 2.6). 
There are better but we dont want to become fancy until we build a certain amount of
tests. Also, we are going to switch to python 3.x in April 2012, and then get a 
more current test module anyway.

"""



import  unittest 


## tools

def run_module(m):
    
    suite_ = unittest.TestLoader().loadTestsFromModule(m)
    runner = unittest.TextTestRunner(verbosity=1)
    runner.run(suite_)

def run_class(c):
    suite = unittest.TestLoader().loadTestsFromTestCase(c)
    runner = unittest.TextTestRunner(verbosity=1)
    runner.run(suite)

def run_name(n):    
    """name n such as 'ascent.Branching.test_ascent_arg'"""
    suite = unittest.TestLoader().loadTestsFromName(n)
    runner = unittest.TextTestRunner(verbosity=1)
    runner.run(suite)


def add_module(suite, testcase):
    suite_ = unittest.TestLoader().loadTestsFromModule(testcase)
    suite.addTest(suite_)



if __name__ == '__main__':
    
    
    import  basics, vcube, ascent, archive, plots, load, semi_uniform, inspection,\
        loop, manipulation

    suite          = unittest.TestSuite()
    add_module(suite, basics)
    add_module(suite, vcube) 
    add_module(suite, ascent)
    add_module(suite, archive)
    add_module(suite, plots)
    add_module(suite, load)
    add_module(suite, semi_uniform)
    add_module(suite, inspection)
    add_module(suite, loop)
    add_module(suite, manipulation)
    
    
    suite_current  = unittest.TestSuite()
    add_module(suite_current, loop)
    
    
    runner = unittest.TextTestRunner(verbosity=1)

    runner.run(suite)
    #runner.run(suite_current)
    
    


